/*
   FINDROM:  A companion program to CODEGEN!

   format of .CGM files:
      char signature[4]
      unsigned int map_ofs
                   =  ofs of start of map in file
      unsigned int load_ofs
      unsigned int load_seg
      unsigned int exe_flag
      unsigned int exe_model
      unsigned int exe_data_seg
      unsigned int flagword
                   CGM_FLAG_RESERVED1 0x0001
                       cgm_flag_domap no longer defined,
                       but reserved for compatibility!
                   CGM_FLAG_INTADDR   0x0002
                   CGM_FLAG_PROCESSOR 0x0004
                   CGM_FLAG_MOREFLAGS 0x8000
      unsigned int start_offset  (always presesnt, bit is reserved)
      unsigned int intnum_offset  (if CGM_FLAG_INTADDR is set)
                   pointer to flagword data, if bit 1 is set.
                   Structure of intaddr data pointed to is:
                      unsigned int num_ints
                      struct
                         {
                         unsigned char intnum
                         unsigned int intofs
                         } int_data[num_ints]
      unsigned int cpu_type  (if CGM_FLAG_PROCESSOR is set)
      CODEGEN map information, crunched as appropriate.


NOTE: In theory, there is 1 word for each flag bit.  What is done with the
      word (info or ptr) is dependent on what the flag represents.  Thus,
      the structure can also be thought of as:
      +------------+-------+-----------------+-----------------+--------------
      | start info | flags | flag data words | additional info | map info...
      +------------+-------+-----------------+-----------------+--------------

*/

#include <compiler.h>
#include <codegen.h>

#define MIN(x, y)   (((x) < (y)) ? (x) : (y))
#define MAX(x, y)   (((x) > (y)) ? (x) : (y))

#define PROG_UNKNOWN    0    /* unknown byte, possibly data */
#define PROG_CODE       1    /* unlabeled Code */
#define PROG_DATA       2    /* Data (labeled, of course!) */
#define PROG_LABEL      4    /* code is labeled */
#define PROG_WORD       8    /* byte or word reference for data */
#define PROG_OFFSET   0x10   /* could be a label */
#define PROG_ADDRESS  0x40   /* location contains a label */
#define PROG_KNOWN    0x80   /* has been disassembled */

#define PROG_CRUNCHED 0x7f  /* crunched for map read/write
                               SOME of these bits have to be exclusive! */

#define ROM_BUFSIZE 8192
#define MAX_ROMS 20

unsigned char copyright_str[] = "FINDROM V1.05  Copyright (c) 1989 by William Couture.\n";

struct cgmap_header cgm_header;  /* will need to write a header for .CGM file */

int num_roms = 0;
int work_rom;  /* rom we are working on */
struct
   {
/*   unsigned int foundseg; */
   unsigned int seg;
   unsigned int ofs;
   unsigned int length;
   unsigned int entry_ofs;
   unsigned int valid_checksum;  /* flag */
   } rom_list[MAX_ROMS];

unsigned int ints_to_rom[NUM_ROM_INTS] = {
0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x40, 0x42, 0x48,
0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77
};  /* interrupts that could point to ROM code */

unsigned int numints;
struct int_offset intlist[NUM_ROM_INTS+0x20];  /* allow for BIOS ints */

unsigned int codesegment;  /* current code segment */
unsigned int datasegment;  /* current data segment */
unsigned int highsegment;  /* the value of the higher of the two segments */
unsigned int mapspace;  /* segment used to hold the byte map of the program */
unsigned int cpu_type = 2;
      /*
         default to 80286 cpu
         0 = 8086
         1 = 80186
         2 = 80286
      */

unsigned int have_entry_point;

char rombuf[ROM_BUFSIZE];  /* need local buffer to hold ROM code */

char globalstr[150];
char filename[150];
char mapname[150];
char nl_str[2] = "\n";  /* used frequently */
char onechar[2] = " ";
char cgm_signature[4] = "cgm";  /* all 4 bytes significant */

char *hexbyte(), *hexword();

char copyright[] = "Copyright (c) 1989 by William Couture.";

unsigned int peekw(seg,ofs)  /* peek a WORD address */
   unsigned int seg,ofs;
   {
   return( PEEK(seg,ofs) + (PEEK(seg,ofs+1) << 8) );
   }

main()
   {
   datasegment = DATASEG;  /* get data and code seg.  Note implied small
                              memory model. */
   codesegment = CODESEG;
   highsegment = datasegment;
   if (codesegment > datasegment)  /* do not assume DS > CS */
      highsegment = codesegment;
   mapspace = highsegment + 0x1000;  /* allow for 64K (small memory model) */
   if (mapspace & 0x0fff)
      mapspace = (mapspace & 0xf000) + 0x1000;
   get_cpu_type();
   find_roms();
   if (num_roms == 0)
      {
      puts(copyright_str);
      puts("Sorry, but I cannot find any ROMs.\n");
      }
   else
      {
      show_roms();
      curpos(24,0);
      }
   }

show_roms()
   {
   unsigned int i, key;

   work_rom = 0;  /* we always have this ROM */
   show_roms_screen();
   do
      {
      if ((key = toupper(ci())) == 0)
         key = tolower(ci());
      switch (key)
         {
         case 'h':
         case 'g':
         case 'i':
         case '-':
            rom_cursor_pos(work_rom,TRUE);
            puts("   ");
            work_rom = MAX(work_rom-1,0);
            rom_cursor_pos(work_rom,TRUE);
            puts("-->");
            break;
         case 'p':
         case 'o':
         case 'q':
         case '+':
         case ' ':
            rom_cursor_pos(work_rom,TRUE);
            puts("   ");
            work_rom = MIN(num_roms-1,work_rom+1);
            rom_cursor_pos(work_rom,TRUE);
            puts("-->");
            break;
         case '\015':
            create_rom_files();
            show_roms_screen();
            break;
         }
      } while (key != ESC_KEY);
   }

show_roms_screen()
   {
   int i;
int j;

   cls();
   puts(copyright_str);
   puts("ROMs are found at:");
   for (i = 0; i < num_roms; i++)
      {
      rom_cursor_pos(i,FALSE);
/*      puts(hexword(&rom_list[i].foundseg,globalstr)); */
      puts(hexword(&rom_list[i].seg,globalstr));
      puts(":");
      puts(hexword(&rom_list[i].ofs,globalstr));
      puts("  ");
      puts(hexword(&rom_list[i].length,globalstr));
      puts("  ");
      j = rom_list[i].length / 1024;
      puts(hexword(&j,globalstr));
      puts("K");
      }
   putsat(23,0,"Use the up and down arrow keys to select a ROM and press Return,\n");
   puts("or press Esc to quit.");
   rom_cursor_pos(work_rom,TRUE);
   puts("-->");
   }

int rom_cursor_pos(num,ispointer)
   int num;
   {
   if (ispointer)
      if (num > 9)
         curpos(num-7,41);
      else
         curpos(num+3,1);
   else
      if (num > 9)
         curpos(num-7,45);
      else
         curpos(num+3,5);
   }

create_rom_files()
   {
   int i, j, whichint, row, temp, romfile, mapfile;
   unsigned int len, numout, startloc, romofs, romseg, intnum_offset;
   long longofs;

   memfill(mapspace,0,0x8000,0);  /* seg, offset, len, char */
   memfill(mapspace+0x800,0,0x8000,0);
   have_entry_point = FALSE;
   romofs = rom_list[work_rom].ofs;
   romseg = rom_list[work_rom].seg;
   numints = 0;
   set_vector_entry_points(romseg,romofs);
   check_bios_entry_points(romseg,romofs);
   if ((!have_entry_point) && (romofs))
      {
/*      romseg = rom_list[work_rom].foundseg; */
      romseg = rom_list[work_rom].seg;
      romofs = 0;
      set_vector_entry_points(romseg,romofs);
      }
   cls();
   puts("ROM at ");
/*   puts(hexword(&rom_list[work_rom].foundseg,globalstr)); */
   puts(hexword(&rom_list[work_rom].seg,globalstr));
   puts("  (");
   puts(hexword(&romseg,globalstr));
   puts(":");
   i = romofs;
   puts(hexword(&i,globalstr));
   puts(" to ");
   puts(hexword(&romseg,globalstr));
   puts(":");
   i += rom_list[work_rom].length;
   puts(hexword(&i,globalstr));
   puts(")\n\n");
   puts("Is this a Hard-Disk controller card ROM (Y/N)? ");
   mygets(globalstr);
   temp = 0;
   if (toupper(globalstr[0]) == 'Y')
      {
      puts("Do you want to try to find the low-level format routine (Y/N)? ");
      mygets(globalstr);
      if (toupper(globalstr[0]) == 'Y')
         {
         temp = find_harddisk_format_routine(romseg,romofs);
         }
      }
   puts("\nInitial ROM entry point ");
   if (rom_list[work_rom].entry_ofs != 0xffff)
      {
      puts("at offset ");
      puts(hexword(&rom_list[work_rom].entry_ofs,globalstr));
      }
   else
      puts("unknown.");
   puts("\nHooked interrupt vectors: ");
   j = FALSE;
   for (i = 0; i < NUM_ROM_INTS; i++)  /* all interrupt numbers */
      {
      whichint = ints_to_rom[i];
      if (check_vector(romseg,romofs,rom_list[work_rom].length,whichint))
         {
         j = TRUE;
         puts(hexbyte(&whichint,globalstr));
         puts(" ");
         }
      }
   if (!j)
      puts("NONE");
   if (temp)
      {
      puts("\nHard disk format routine at offset ");
      puts(hexword(&temp,globalstr));
      }
   puts(nl_str);
   if ((rom_list[work_rom].entry_ofs == 0xffff) && (have_entry_point == FALSE))
      {
      puts("\nI cannot find any entry points for this ROM.\n");
      puts("Do you want to continue (Y/N)? ");
      mygets(globalstr);
      if (toupper(globalstr[0]) != 'Y')
         return;
      }
   get_curpos(&row,&temp);  /* only row will be used */
   row++;
   while (1)  /* until I'm satisfied */
      {
      clear_to_bottom(row,0);
      strcpy(filename,"ROM_");
      hexword(&rom_list[work_rom].seg,globalstr);
      strcat(filename,globalstr);
      strcat(filename,".DAT");
      putsat(row,0,"Enter the ROM filename, return for\n");
      puts(filename);
      puts(", or Esc to quit? ");
      if (mygets(globalstr) == -1)
         return;
      if (globalstr[0])
         strcpy(filename,globalstr);
      if ((romfile = open(filename,0)) != -1)  /* 0 = read mode */
         {
         close(romfile);
         putsat(row+3,0,"The file ");
         puts(filename);
         puts("\nalready exists.  Do you want to over-write it (Y/N, Esc = quit)? ");
         if (mygets(globalstr) == -1)
            return;
         if (toupper(globalstr[0]) == 'Y')
            break;
         }
      else
         break;
      }
   get_curpos(&row,&temp);  /* only row will be used */
   row++;
   while (1)  /* until I'm satisfied */
      {
      clear_to_bottom(row,0);
      strcpy(mapname,filename);
      for (i = strlen(mapname); i; i--)
         if (mapname[i] == '.')
            mapname[i] = '\0';
      strcat(mapname,".CGM");
      putsat(row,0,"Enter the Codegen Map filename, return for\n");
      puts(mapname);
      puts(", or Esc to quit? ");
      if (mygets(globalstr) == -1)
         return;
      if (globalstr[0])
         strcpy(mapname,globalstr);
      if ((romfile = open(mapname,0)) != -1)  /* 0 = read mode */
         {
         close(romfile);
         putsat(row+3,0,"The file ");
         puts(mapname);
         puts("\nalready exists.  Do you want to over-write it (Y/N, Esc = quit)? ");
         if (mygets(globalstr) == -1)
            return;
         if (toupper(globalstr[0]) == 'Y')
            break;
         }
      else
         break;
      }
   len = rom_list[work_rom].length;
   numout = romofs;
   romfile = creat(filename);
   do
      {
      i = MIN(ROM_BUFSIZE,len);
      len -= i;
/*      MEMCPY(i,rom_list[work_rom].foundseg,numout,datasegment,rombuf); */
      MEMCPY(i,romseg,numout,datasegment,rombuf);
      numout += i;
      if (write(romfile,rombuf,i) != i)
         {
         close(romfile);
         unlink(filename);
         puts("\nError writing ROM file.  Check for disk full or write protect.\n");
         puts("press any key to continue...");
         ci();
         return;
         }
      } while (len);
   close(romfile);
   mapfile = creat(mapname);
   strcpy(cgm_header.signature,cgm_signature);
   cgm_header.map_ofs = sizeof(struct cgmap_header) + 4;
   cgm_header.load_ofs = romofs;
   cgm_header.load_seg = 0;  /* for future expansion */
   cgm_header.filetype_flag = 0;  /* COM style (1 = EXE, 2 = device) */
   cgm_header.exe_model = 0;  /* .COM memory model */
   cgm_header.exe_data_seg = 0;
   cgm_header.flagword = CGM_FLAG_PROCESSOR;  /* processor type is defined */
   intnum_offset = 0;
   if (numints)
      {
      cgm_header.flagword |= CGM_FLAG_INTADDR;
      intnum_offset = 4 + sizeof(struct cgmap_header);
      cgm_header.map_ofs += 2 + numints * sizeof(struct int_offset);
      }
   startloc = rom_list[work_rom].entry_ofs;
   if (startloc == 0xffff)
      for (startloc = 0; startloc < 0xffff; startloc++)
         if (PEEK(mapspace,startloc) == (PROG_CODE | PROG_LABEL))
            break;
   cgm_header.start_offset = startloc;
   write(mapfile,&cgm_header,sizeof(struct cgmap_header));
   write(mapfile,&intnum_offset,2);
   write(mapfile,&cpu_type,2);
   if (numints)
      {
      lseek(mapfile,(long)intnum_offset,0);
      write(mapfile,&numints,2);
      write(mapfile,intlist,numints * sizeof(struct int_offset));
      }
   longofs = 0;
   while (len = crunch_it(mapspace,&longofs))
      {
      if (write(mapfile,rombuf,len) != len)
         {
         close(mapfile);
         unlink(mapname);
         puts("\nError writing Map file.  Check for disk full.\n");
         puts("press any key to continue...");
         ci();
         return;
         }
      }
   close(mapfile);
   }

unsigned int is_copyright_rom(seg)
   unsigned int seg;
   {
   int i, j;

   if ((peekw(seg,0) == 0xaa55) || (peekw(seg,0) == 0x55aa))
      return(0);  /* assume can find expansion roms normally */
   j = FALSE;
   MEMCPY(512,seg,0,datasegment,rombuf);
   for (i = 0; i < 256; i++)
      if ((strncmpi("copyright",&rombuf[i],9) == 0)
                   || (strncmpi("ccooppyyrriigghhtt",&rombuf[i],18) == 0))
         {
         j = TRUE;
         break;
         }
   if (!j)
      for (i = 0; i < 256; i++)
         if ((strncmpi("(c)",&rombuf[i],3) == 0)
                || (strncmpi("((cc))",&rombuf[i],6) == 0))  /* find it the hard way */
            {
            j = TRUE;
            break;
            }
   if (!j)
      for (i = 0; i < 256; i++)
         if ((strncmpi("copr.",&rombuf[i],5) == 0)
                || (strncmpi("ccoopprr..",&rombuf[i],10) == 0))  /* A Harder way! */
            {
            j = TRUE;
            break;
            }
   if (!j)
      return(0);
   if (seg & 0xfff)
      return(0x7fff);
   return(0xffff);
   }

find_roms()
   {
   unsigned int i, j, search_seg, check_length, checksum;

   if (i = is_copyright_rom(0xfe00))
      {
/*      rom_list[0].foundseg = 0xfe00; */
      rom_list[0].seg = 0xf000;
      rom_list[0].ofs = 0xe000;
      rom_list[0].length = 0x2000;
      rom_list[0].entry_ofs = 0xfff0;
      rom_list[0].valid_checksum = TRUE;
      num_roms = 1;
      }
   if (i = is_copyright_rom(0xf000))
      {
/*      rom_list[0].foundseg = */
      rom_list[0].seg = 0xf000;
      rom_list[0].ofs = 0;
      rom_list[0].length = 0xffff;
      rom_list[0].entry_ofs = 0xfff0;
      rom_list[0].valid_checksum = TRUE;
      num_roms = 1;
      }
   else
      {
      if (i = is_copyright_rom(0xf800))
            {
/*            rom_list[0].foundseg = 0xf800; */
            rom_list[0].seg = 0xf000;
            rom_list[0].ofs = 0x8000;
            rom_list[0].length = 0x7fff;
            rom_list[0].entry_ofs = 0xfff0;
            rom_list[0].valid_checksum = TRUE;
            num_roms = 1;
            }
      }
   for (search_seg = 0xc000; (search_seg <= 0xf400) && (num_roms < MAX_ROMS);
                                                       search_seg += 0x400)
      if ((peekw(search_seg,0) == 0xaa55) || (peekw(search_seg,0) == 0x55aa))
         {
         check_length = PEEK(search_seg,2) * 512;
/*         rom_list[num_roms].foundseg = */
         rom_list[num_roms].seg = search_seg;
         rom_list[num_roms].ofs = 0;
         rom_list[num_roms].length = check_length;
         rom_list[num_roms].entry_ofs = 3;
         checksum = 0;
         do
            {
            checksum += PEEK(search_seg,--check_length);
            } while (check_length);
         rom_list[num_roms++].valid_checksum = checksum ? FALSE : TRUE;
         }
   for (search_seg = 0xc000; (search_seg < 0xf000) && (num_roms < MAX_ROMS);
                                            search_seg += 0x800)
      {
      j = TRUE;  /* is a unique segment */
      for (i = 0; i < num_roms; i++)
         if ((search_seg == rom_list[i].seg)
                || ((search_seg > rom_list[i].seg)
                        && (search_seg < (rom_list[i].seg+(rom_list[i].length >> 4)))))
            j = FALSE;
      if (j)
         {
         if (i = is_copyright_rom(search_seg))
            {
/*            rom_list[num_roms].foundseg = */
            rom_list[num_roms].seg = search_seg;
            rom_list[num_roms].ofs = 0;
            rom_list[num_roms].length = i;
            rom_list[num_roms].entry_ofs = 0xffff;  /* flag: don't know entry point */
            rom_list[num_roms++].valid_checksum = TRUE;
            }
         }
      }
   }

int set_vector_entry_points(seg,ofs)
   unsigned int seg, ofs;
   {
   int i,whichint;
   unsigned int intseg, intofs;

   for (i = 0; i < NUM_ROM_INTS; i++)  /* all interrupt numbers */
      {
      whichint = ints_to_rom[i];
      if (check_vector(seg,ofs,rom_list[work_rom].length,whichint))
         {
         intseg = peekw(0,whichint*4+2);
         intofs = peekw(0,whichint*4);
         if (seg > intseg)
            intofs -= (seg-intseg) * 16;
         else
            intofs += (intseg-seg) * 16;
         POKE(PROG_CODE | PROG_LABEL,mapspace,intofs);
         have_entry_point = TRUE;
         intlist[numints].intnum = whichint;
         intlist[numints].intofs = intofs;
         numints++;
         }
      }
   }

check_bios_entry_points(seg,ofs)
   unsigned int seg, ofs;
   {
   unsigned int i, j, whichint, match, intseg, intofs;
   long romstart, romend, jmpdest;

   romstart = ((long)seg << 4) + ofs;
   romend = romstart + rom_list[work_rom].length;
   if ((romstart < 0xffef3L) && (romend > 0xfff23L))  /* if includes BIOS table */
      {
      match = TRUE;
      for (i = 0; i < NUM_ROM_INTS; i++)  /* all interrupt numbers */
         {
         whichint = ints_to_rom[i];
         if ((whichint >= 8) && (whichint < 0x1f))
            if ((check_vector(seg,ofs,rom_list[work_rom].length,whichint)) &&
                  (peekw(0,whichint*4) != peekw(0xf000,0xfef3+(whichint-8)*2))
                  && (peekw(0xf000,0xfef3+(whichint-8)*2) != 0x0000))
               match = FALSE;
         }
      if (match)
         {
         have_entry_point = TRUE;
         for (whichint = 8; whichint < 0x1d; whichint++)
            {
            intofs = peekw(0xf000,0xfef3+(whichint-8)*2);
            if (intofs == 0x0000)
               continue;  /* unused, skip it! */
            if (seg > 0xf000)
               intofs -= (seg - 0xf000) * 16;
            else
               intofs += (0xf000 - seg) * 16;
            POKE(PROG_CODE | PROG_LABEL,mapspace,intofs);
            for (j = 0; j < numints; j++)
               if (intlist[j].intnum == whichint)
                  break;
            if (j == numints)  /* if this int not found */
               {
               intlist[numints].intnum = whichint;
               intlist[numints].intofs = intofs;
               numints++;
               }
            }
         }
      }
   if ((romstart <= 0xffff0L) && (romend >= 0xffff6L))  /* if includes reset */
      {
      if (PEEK(0xffff,0) == 0xea)  /* if JMP SEG:OFS */
         {
         intseg = peekw(0xffff,3);
         intofs = peekw(0xffff,1);
         jmpdest = ((long)intseg << 4) + intofs;
         if ((jmpdest >= romstart) && (jmpdest <= romend))
            {
            if (seg > intseg)
               intofs -= (seg-intseg) * 16;
            else
               intofs += (intseg-seg) * 16;
            POKE(PROG_CODE | PROG_LABEL,mapspace,intofs);
            have_entry_point = TRUE;
            }
         }
      }
   }

int check_vector(seg,ofs,length,interrupt)
   unsigned int seg, ofs, length, interrupt;
   {
   unsigned int pointer;

   seg += ofs >> 4;
   pointer = (peekw(0,interrupt*4) >> 4) + peekw(0,interrupt*4+2);
   if ((pointer >= seg) && (pointer < (seg + (length >> 4))))
      return(TRUE);
   return(FALSE);
   }

find_harddisk_format_routine(seg,ofs)
   unsigned int seg, ofs;
   {
   unsigned int i, j;

   if ((PEEK(seg,ofs+3) == 0xeb)
                   && (((i=PEEK(seg,ofs+5)) == 0xe9) || (i == 0xeb)))
/*
   i.e., format is:
      JMP short init_code
      JMP format_routine
*/
      {
      POKE(PROG_CODE | PROG_LABEL,mapspace,ofs+5);
      have_entry_point = TRUE;
      return(ofs+5);
      }
   if ((PEEK(seg,ofs+3) == 0xe9)
                   && (((i=PEEK(seg,ofs+6)) == 0xe9) || (i == 0xeb)))
/*
   i.e., format is:
      JMP init_code
      JMP format_routine
*/
      {
      POKE(PROG_CODE | PROG_LABEL,mapspace,ofs+6);
      have_entry_point = TRUE;
      return(ofs+6);
      }
   MEMCPY(512,seg,0,datasegment,rombuf);  /* see if we recognize who
                                             made this ROM */
   j = FALSE;
   for (i = 0; i < 256; i++)
      if (strncmpi("copyright",&rombuf[i],9) == 0)
         j = TRUE;
   if (!j)
      return(0);  /* can't find copyright notice, don't keep looking */
   for (i = 0; i < 256; i++)
      if (strncmpi("western digital",&rombuf[i],15) == 0)
         {
         POKE(PROG_CODE | PROG_LABEL,mapspace,ofs+5);
         have_entry_point = TRUE;
         return(ofs+5);
         }
   for (i = 0; i < 256; i++)
      if (strncmpi("seagate",&rombuf[i],7) == 0)
         {
         POKE(PROG_CODE | PROG_LABEL,mapspace,ofs+5);
         have_entry_point = TRUE;
         return(ofs+5);
         }
   for (i = 0; i < 256; i++)
      if (strncmpi("adaptec",&rombuf[i],7) == 0)
         {
         POKE(PROG_CODE | PROG_LABEL,mapspace,ofs+0xccc);  /* FUNKY! */
         have_entry_point = TRUE;
         return(ofs+0xccc);
         }
   for (i = 0; i < 256; i++)
      if (strncmpi("sms/omti",&rombuf[i],8) == 0)
         {
         POKE(PROG_CODE | PROG_LABEL,mapspace,ofs+6);
         have_entry_point = TRUE;
         return(ofs+6);
         }
   for (i = 0; i < 256; i++)
      if (strncmpi("omti",&rombuf[i],4) == 0)  /* sms/omti */
         {
         POKE(PROG_CODE | PROG_LABEL,mapspace,ofs+6);
         have_entry_point = TRUE;
         return(ofs+6);
         }
   for (i = 0; i < 256; i++)
      if (strncmpi("wdc",&rombuf[i],3) == 0)  /* Western digital again */
         {
         POKE(PROG_CODE | PROG_LABEL,mapspace,ofs+5);
         have_entry_point = TRUE;
         return(ofs+5);
         }
   for (i = 0; i < 256; i++)
      if (strncmpi("sms",&rombuf[i],3) == 0)  /* sms/omti */
         {
         POKE(PROG_CODE | PROG_LABEL,mapspace,ofs+6);
         have_entry_point = TRUE;
         return(ofs+6);
         }
   }

void putsat(r,c,s)
   int r,c;
   char *s;
   {
   curpos(r,c);
   puts(s);
   }

memfill(seg,offset,len,byte)
     /* +4  +6     +8  +10 */
   {
#asm
   mov es,word [bp+4]
   mov di,word [bp+6]
   mov cx,word [bp+8]
   mov al,byte [bp+10]
   cld                 ;up!
   db 0f3h             ;REP
   stosb               ;fill with byte
#endasm
   }

void cls()
   {
#asm
   mov cx,0     ;row,col of upper left corner
   mov dx,184fh ;row,col of lower right corner
   mov bh,7     ;fill attribute
   mov ax,600h  ;scroll up entire window
   push bp      ;save bp
   int 10h      ;call BIOS
   pop bp       ;restore bp
   mov dx,0     ;row 0, col 0
   mov bh,0     ;page 0
   mov ah,2     ;set cursor position
   push bp      ;save bp
   int 10h      ;call BIOS
   pop bp       ;restore bp
#endasm
   }

clear_to_bottom(r,c)
   {
#asm
   mov ch,byte [bp+4]   ;upper left row
   mov cl,byte [bp+6]   ;upper left col
   mov dh,ch            ;lower right row
   mov dl,79            ;lower right col
   mov bh,7             ;fill attribute
   mov ax,600h          ;scroll up and erase
   push cx              ;save upper left corner
   push bp              ;save bp
   int 10h              ;call BIOS
   pop bp               ;restore bp
   pop cx               ;restore upper left corner
   cmp ch,18h           ;already at end?
   je end_of_clear_to_bottom  ;yes, we're done
   inc ch               ;next row
   xor cl,cl            ;start at col 0
   mov dx,184fh         ;lower right is row 24, col 79
   mov bh,7             ;fill attribute
   mov ax,600h          ;scroll up entire region
   push bp              ;save BP
   int 10h              ;call BIOS
   pop bp               ;restore BP
end_of_clear_to_bottom:
#endasm
   }

void curpos(row,col)
   int row,col;
   {
#asm
   mov dh,byte [bp+4]  ;row
   mov dl,byte [bp+6]  ;col
   mov bh,0            ;page
   mov ah,2            ;set cursor position
   push bp             ;save bp
   int 10h             ;call BIOS
   pop bp              ;restore bp
#endasm
   }

gchar()
   {
#asm
   mov dl,0ffh
   mov ah,6
   int 21h
   jz  gchar1
   mov ah,0
   jmp gchar2
gchar1: mov ah,1
gchar2:
#endasm
   }

char hexdig(ch)  /* return the ASCII char corresponding to the HEX value */
   char ch;
   {
   return(ch > '\011' ? (ch-'\012' + 'A') : (ch+'0'));
   }

char *hexbyte(src,target)  /* encode the first two bytes of src into target
                              as ascii, and return the address of target.
                              Note that this is sometimes called with the
                              form &intvar for src.  THIS ONLY WORKS ON A
                              MACHINE THAT USES LOW BYTE FIRST!!
                              Sorry about that if you want to port it, but
                              I was lazy. */
   char *src, *target;
   {
   target[0] = hexdig( (src[0] >> 4) & 0x0f);
   target[1] = hexdig(src[0] & 0x0f);
   target[2] = '\0';
   return(target);
   }

char *hexword(src,target)  /* As for hexbyte, but for a two byte word.
                              The comment about being called with
                              &intvar for src applies DOUBLY here! */
   char *src, *target;
   {
   target[0] = hexdig( (src[1] >> 4) & 0x0f);
   target[1] = hexdig(src[1] & 0x0f);
   target[2] = hexdig( (src[0] >> 4) & 0x0f);
   target[3] = hexdig(src[0] & 0x0f);
   target[4] = '\0';
   return(target);
   }

int strncmpi(s1, s2, n)
   char *s1, *s2;
   int n;
   {
   while ((n--) && (*s1) && (*s2))  /* while count and not EOS */
      {
      if (tolower(*s1) != tolower(*s2))  /* if different, say so */
         return(tolower(*s1) - tolower(*s2));
      s1++;  /* else next char */
      s2++;
      }
   if (n != -1)  /* if stopped by EOS, make sure both have EOS */
      return(*s1 - *s2);
   return(0);
   }

get_curpos(row,col)
   int *row, *col;
   {
#asm
   mov bh,0           ;page 0
   mov ah,3           ;fn = get cursor position
   push bp            ;keep BP safe from BIOS
   int 10h            ;call video BIOS
   pop bp             ;restore bp
   xor ax,ax          ;ax = 0
   mov al,dh          ;ax = row position of cursor
   mov bx,word [bp+4] ;get addr of ROW
   mov word [bx],ax   ;set row
   mov al,dl          ;ax = col position of curosr
   mov bx,word [bp+6] ;get addr of COL
   mov word [bx],ax   ;set col
#endasm
   }

int mygets(str)
   char *str;
   {
   int len,ch,row,col,col2,temp,temp2;

   get_curpos(&row,&col);  /* get cursor row */
   len = 0;
   while (1)             /* forever */
      {
      do
         {
         while ((ch = gchar()) > 255)
            ;                  /* wait for keypress */
         if (!ch)              /* if null (1st char of extended keypress) */
            while (gchar() > 255)
               ;
         } while (!ch);
      if (ch == ESC_KEY)
         {
         str[0] = '\0';
         if (row < 24)
            puts(nl_str);
         return(-1);
         }
      if (ch == '\010')
         {
         if (len)
            {
            len--;
            if (str[len] == '\t')
               {
               temp2 = col;
               for (temp = 0; temp < len; temp++)
                  temp2 += (str[temp] == '\t') ? (8-temp2%8) : 1;
               get_curpos(&temp,&col2);
               for ( ; temp2 < col2; temp2++)
                  puts("\010");
               }
            else
               puts("\010 \010");
            }
         }
      else
         if (ch == '\015')
            {
            str[len] = '\0';
            if (row < 24)
               puts(nl_str);
            return(len);
            }
         else
            {
            get_curpos(&temp,&col2);
            temp = col2 + 8 - col2 % 8;
            if (ch == '\t')
               {
               if (temp > 78)
                  putchar('\007');
               else
                  {
                  str[len++] = ch;
                  onechar[0] = ' ';
                  for (temp -= col2; temp; temp--)
                     puts(onechar);
                  }
               }
            else
               if ((len < 70) && (col2+1 < 78))
                  {
                  if ((ch >= 32) && (ch < 128))
                     {
                     str[len++] = ch;
                     onechar[0] = ch;
                     puts(onechar);
                     }
                  }
               else
                  putchar('\007');
            }
      }
   }

int crunch_it(seg,ofs)
   unsigned int seg;
   long *ofs;
   {
   long count, count3;
   unsigned int repeat_len, count2;
   unsigned char ch;

   count = *ofs;
   count2 = 0;
   while ((count <= 65535) && (count2 < ROM_BUFSIZE-4))  /* make sure we do not overflow buffer */
      {
      ch = PEEK(seg,(unsigned int)count);
      repeat_len = 1;  /* can count the 1st char as a repeat, too */
      for (count3 = count+1; (count3 < 65535) && (PEEK(seg,(unsigned int)count3) == ch);
                                                                  count3++)
         repeat_len++;
      if (repeat_len > 4)
         {
         rombuf[count2++] = PROG_CRUNCHED;
         *(unsigned int *)&rombuf[count2] = repeat_len;
         count2 += 2;
         count += repeat_len - 1;  /* -1 to offset the count++ below */
         }
      rombuf[count2++] = ch;
      count++;
      };
   *ofs = count;
   return(count2);
   }

get_cpu_type()
   {
#asm
   push sp              ;our sp
   pop  ax              ;to ax
   cmp  ax,sp           ;the same, is an 80286
   je   have_cpu_type   ;yes, we have it
   dec  word cpu_type_  ;try 80186
   mov  al,0ffh         ;set all bits of al
   mov  cl,33           ;80186 masks shifts mod 32
   shr  al,cl           ;see if anything left
   or   al,al           ;all bits cleared?
   jnz  have_cpu_type   ;no, it's an 80186
   dec  word cpu_type_  ;else its 8086
have_cpu_type:
#endasm
   }
